#ifndef bqtExprHH
#define bqtExprHH

#include <string>
#include <vector>

#include "autoptr"
#include "tristate"

class expr: public ptrable
{
public:
    expr();
    virtual ~expr();
    
    virtual tristate is_num() const = 0;
    virtual tristate is_bool() const = 0;
    virtual tristate is_string() const = 0;
    virtual bool is_const() const = 0;
};
typedef autoptr<expr> exprp;

class expressions: public std::vector<exprp>
{
public:
    expressions() { }
	expressions(expr *b)    { push_back(b); }
};

#include "qb.tab.h"

class expr_unary: public expr
{
    int token;
    exprp p;
public:
	expr_unary(int tok, expr *a) : token(tok), p(a) { }

    virtual tristate is_num() const    { return p->is_num(); }
    virtual tristate is_bool() const   { return p->is_bool(); }
    virtual tristate is_string() const { return p->is_string(); }
    virtual bool is_const() const  { return p->is_const(); }
};
class expr_binary: public expr
{
    int token; 
    exprp p0, p1;
public:
	expr_binary(expr *a, int tok, expr *b) : token(tok), p0(a), p1(b) { }

    virtual tristate is_num() const    { return maybe; } //FIXME
    virtual tristate is_bool() const   { return maybe; }
    virtual tristate is_string() const { return maybe; }
    virtual bool is_const() const  { return p0->is_const() && p1->is_const(); }
};

class expr_number: public expr
{
    enum { t_int, t_float } type;
    long long int int_val;
    long double float_val;
    // TODO: value here.
public:
    expr_number() : type(t_int), int_val(0), float_val(0) { }
    expr_number(long long int i) : type(t_int), int_val(i) { }
    expr_number(long int i) : type(t_int), int_val(i) { }
    expr_number(long double f) : type(t_float), float_val(f) { }
    expr_number(int i) : type(t_int), int_val(i) { }
    expr_number(double f) : type(t_float), float_val(f) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const  { return true; }
};

class expr_string: public expr
{
    std::string value;
public:
    expr_string() { }
    expr_string(const char* v) : value(v) { }
    expr_string(const std::string& v) : value(v) { }

    virtual tristate is_num() const    { return false; }
    virtual tristate is_bool() const   { return false; }
    virtual tristate is_string() const { return true; }
    virtual bool is_const() const  { return true; }
};

class expr_userfun: public expr
{
    std::string name;
    expressions params;
public:
    expr_userfun(const std::string& n) : name(n) { }    
    
    virtual tristate is_num() const    { return maybe; } // FIXME
    virtual tristate is_bool() const   { return maybe; } // FIXME
    virtual tristate is_string() const { return maybe; } // FIXME
    virtual bool is_const() const  { return false; }

    void set_params(expressions* ind)
    {
    	params = *ind;
    	delete ind;
    }
};

class expr_builtin_numeric: public expr
{
    int token;
    expressions params;
public:
    expr_builtin_numeric(int tok) : token(tok) { }
    expr_builtin_numeric(int tok, expr* a) : token(tok)
        { params.push_back(a); }
    expr_builtin_numeric(int tok, expr* a, expr* b) : token(tok)
        { params.push_back(a); params.push_back(b); }
    expr_builtin_numeric(int tok, expr* a, expr* b, expr* c) : token(tok)
        { params.push_back(a); params.push_back(b); params.push_back(c); }
    expr_builtin_numeric(int tok, expr* a, expr* b, expr* c, expr* d) : token(tok)
        { params.push_back(a); params.push_back(b); params.push_back(c); params.push_back(d); }


    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const
    {
        for(unsigned a=0; a<params.size(); ++a)
            if(!params[a]->is_const()) return false; 
        return (token == ABS
             || token == ATN
             || token == CDBL
             || token == CINT
             || token == CLNG
             || token == COS
             || token == CSNG
             || token == EXP
             || token == FIX
             || token == INT
             || token == LOG
             || token == SQR
             || token == TAN
             || token == VAL
               );
    }
};

class expr_builtin_string: public expr
{
    int token;
    expressions params;
public:
    expr_builtin_string(int tok) : token(tok) { }
    expr_builtin_string(int tok, expr *a) : token(tok)
      { params.push_back(a); }
    expr_builtin_string(int tok, expr *a, expr *b) : token(tok)
      { params.push_back(a); params.push_back(b); }
    expr_builtin_string(int tok, expr *a, expr *b, expr *c) : token(tok)
      { params.push_back(a); params.push_back(b); params.push_back(c); }
    expr_builtin_string(int tok, expr *a, expr *b, expr *c, expr *d) : token(tok)
      { params.push_back(a); params.push_back(b);
        params.push_back(c); params.push_back(d); }
    
    virtual tristate is_num() const    { return false; }
    virtual tristate is_bool() const   { return false; }
    virtual tristate is_string() const { return true; }
    virtual bool is_const() const
    {
        for(unsigned a=0; a<params.size(); ++a)
            if(!params[a]->is_const()) return false; 
        
        /* Pure functions: functions that only depend on their parameter;
         * can be calculated at compile-time. */
        return (token == LTRIM_S
             || token == RTRIM_S
             || token == LCASE_S
             || token == UCASE_S
             || token == INSTR
             || token == MID_S
             || token == RIGHT_S
             || token == STRING_S
             || token == CHR_S
             || token == HEX_S
             || token == MKD_S
             || token == MKI_S
             || token == MKL_S
             || token == MKS_S
             || token == OCT_S
             || token == SPACE_S
             || token == STR_S
               );
    }
};

class expr_coordinate: public expr
{
    bool step;
    exprp x, y;
public:
    virtual tristate is_num() const    { return false; }
    virtual tristate is_bool() const   { return false; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const { return x->is_const() && y->is_const(); }
};

class expr_var: public expr
{
    std::string name;
    expressions indices;
    autoptr<expr_var> member;            // NULL if none
public:
    expr_var(const std::string& n) : name(n) { }    
    
    virtual tristate is_num() const    { return maybe; } // FIXME
    virtual tristate is_bool() const   { return maybe; } // FIXME
    virtual tristate is_string() const { return maybe; } // FIXME
    virtual bool is_const() const { return false; }
    
    void set_indices(expressions* ind)
    {
    	indices = *ind;
    	delete ind;
    }
    void set_ptr(expr_var *b)
    {
    	member = b;
    }
};

class expr_lbound: public expr
{
    std::string name;
    exprp num_index;
public:
    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }
};

class expr_ubound: public expr
{
    std::string name;
    exprp num_index;
public:
    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }
};

#endif
